from datetime import datetime, timedelta, timezone

from livestyled.models.fixture import Fixture
from livestyled.schemas.fixture import FixtureSchema


def test_create_fixture_from_deserialized():
    deserialized_fixture = {
        'away_id': 61,
        'away_score': {'goals': 1, 'penalties': None},
        'competition_id': 6,
        'home_id': 6,
        'home_score': {'goals': 1, 'penalties': None},
        'id': 198,
        'is_fulltime': False,
        'is_terminated': False,
        'season_id': 2,
        'start_at': datetime(2020, 8, 28, 9, 46, 2, tzinfo=timezone(timedelta(0), '+0000')),
        'venue_id': 5,
        'external_id': None,
        'status': 'Active',
        'url': {
            'url': 'https://test.com',
            'is_enabled': True,
            'title': 'TEST'
        }
    }
    fixture = Fixture(**deserialized_fixture)
    assert fixture


def test_serialize_fixture():
    deserialized_fixture = {
        'away_id': 61,
        'away_score': {'goals': 1, 'penalties': None},
        'competition_id': 6,
        'home_id': 6,
        'home_score': {'goals': 1, 'penalties': None},
        'id': 198,
        'is_fulltime': False,
        'is_terminated': False,
        'season_id': 2,
        'start_at': datetime(2020, 8, 28, 9, 46, 2, tzinfo=timezone(timedelta(0), '+0000')),
        'venue_id': 5,
        'external_id': None,
        'status': 'Active',
        'url': {
            'url': 'https://test.com',
            'is_enabled': True,
            'title': 'TEST'
        }
    }
    fixture = Fixture(**deserialized_fixture)
    serialized_fixture = FixtureSchema().dump(fixture)
    assert serialized_fixture
    assert serialized_fixture['competition'] == '/v4/competitions/6'
    assert serialized_fixture['url'] == {
        'url': 'https://test.com',
        'is_enabled': True,
        'title': 'TEST'
    }



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.product import (
    Product,
    ProductCategory,
    ProductImage,
    ProductTranslation,
    ProductVariant
)
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.fulfilment_point import FulfilmentPointSchema


class ProductVariantSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = ProductVariant

    id = fields.Int()
    price = fields.Integer()
    stocks = fields.List(fields.String)


class ProductCategorySchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = ProductCategory
        api_type = 'product_categories'
        url = 'v4/product_categories'

    id = fields.Int()
    reference = fields.String(missing=None)
    position = fields.Int()


class ProductImageSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = ProductImage

    id = fields.Integer()
    product = RelatedResourceLinkField(schema='livestyled.schemas.order.ProductSchema')
    position = fields.Integer()
    image_url = fields.String(data_key='imageUrl', missing=None)
    external_id = fields.String(data_key='externalId', missing=None)

    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True, missing=None)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True, missing=None)


class ProductTranslationSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = ProductTranslation

    id = fields.Int()
    language = fields.String()
    title = fields.String(missing=None)
    description = fields.String(missing=None)

    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True, missing=None)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True, missing=None)


class CoreProductCategorySchema(Schema):
    class Meta:
        unknown = EXCLUDE
    id = fields.Integer()
    name = fields.String()


class ProductSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = Product
        api_type = 'products'
        url = 'v4/products'

    id = fields.Int()
    status = fields.String(missing=None)
    reference = fields.String(missing=None)
    modifier_lists = fields.Raw(missing=[], data_key='modifierLists')
    external_id = fields.String(data_key='externalId', missing=None)
    holding_time = fields.Raw(missing=None, data_key='holdingTime')
    reconciliation_group = fields.Raw(missing=None, data_key='reconciliation_group')
    images = RelatedResourceField(schema=ProductImageSchema, many=True)
    categories = RelatedResourceField(schema=ProductCategorySchema, many=True, missing=[])
    translations = RelatedResourceField(schema=ProductTranslationSchema, many=True, missing=[])
    fulfilment_points = RelatedResourceLinkField(schema=FulfilmentPointSchema, many=True, missing=[], data_key='fulfilmentPoints')
    variants = RelatedResourceLinkField(schema=ProductVariantSchema, many=True, missing=[])
    core_product_category = fields.Nested(CoreProductCategorySchema, missing=None, data_key='coreProductCategory')



class News:
    def __init__(
            self,
            id: int or None,
            headline: str or None,
            url: str,
            author: str,
            image_url: str or None,
            title: str or None,
            external_id: str or None,
            media: dict or None,
            published_at,
            updated_at,
    ):
        self.id = id
        self.headline = headline
        self.image_url = image_url
        self.title = title
        self.external_id = external_id
        self.media = media
        self.url = url
        self.author = author
        self.published_at = published_at
        self.updated_at = updated_at

    @classmethod
    def create_new(
            cls,
            external_id: str,
            headline: str,
            image_url: str,
            title: str,
            media: dict or None,
            url: str,
            author: str,
            published_at,
            updated_at,
    ):
        return News(
            id=None,
            headline=headline,
            image_url=image_url,
            title=title,
            media=media,
            external_id=external_id,
            url=url,
            author=author,
            published_at=published_at,
            updated_at=updated_at,
        )

    def diff(self, other):
        differences = {}
        fields = (
            'headline', 'image_url', 'title', 'external_id', 'media', 'url', 'author', 'published_at',
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.ticket_auth import TicketAuth
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.ticket_integration import TicketIntegrationSchema
from livestyled.schemas.user import UserEmailSchema


class TicketAuthSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'ticket_auths'
        url = 'v4/ticket_auths'
        model = TicketAuth

    id = fields.Int()
    client_email = fields.String(required=False, missing=None, data_key='clientEmail')
    client_id = fields.String(required=False, missing=None, data_key='clientId')
    access_token = fields.String(required=False, missing=None, data_key='accessToken')
    expire_at = fields.AwareDateTime(data_key='expireAt', allow_none=False, missing=None)
    refresh_token = fields.String(required=False, missing=None, data_key='refreshToken')
    user_email = RelatedResourceField(schema=UserEmailSchema, required=False, missing=None, data_key='userEmail')
    ticket_integration = RelatedResourceLinkField(schema=TicketIntegrationSchema, required=False, missing=None, data_key='ticketIntegration')



from collections import namedtuple
from datetime import datetime

from livestyled.models.competition import Competition
from livestyled.models.season import Season
from livestyled.models.sport_venue import SportVenue
from livestyled.models.team import Team


Score = namedtuple('Score', 'goals, penalties')
Url = namedtuple('Url', 'title, url, is_enabled')


class Fixture:
    def __init__(
            self,
            id,
            external_id,
            start_at,
            is_fulltime,
            is_terminated,
            home_id,
            away_id,
            home_score,
            away_score,
            season_id,
            competition_id,
            venue_id,
            status,
            url,
            allow_overwrite=False
    ):
        self._id = id
        self._external_id = external_id
        self._start_at = start_at
        self._is_fulltime = is_fulltime
        self._is_terminated = is_terminated
        self._home = Team.placeholder(id=home_id)
        self._away = Team.placeholder(id=away_id)
        self._home_goals = home_score['goals']
        self._home_penalties = home_score['penalties']
        self._away_goals = away_score['goals']
        self._away_penalties = away_score['penalties']
        self._season = Season.placeholder(id=season_id)
        self._competition = Competition.placeholder(id=competition_id)
        self._venue = SportVenue.placeholder(id=venue_id)
        self._status = status
        if url:
            if isinstance(url, Url):
                self.url = url
            elif isinstance(url, dict):
                self.url = Url(**url)
        else:
            self.url = None
        self.allow_overwrite = allow_overwrite

    @classmethod
    def create_new(
            cls,
            external_id: str,
            start_at: datetime,
            is_fulltime: bool,
            is_terminated: bool,
            home: Team or str or int,
            away: Team or str or int,
            home_goals: int or None,
            home_penalties: int or None,
            away_goals: int or None,
            away_penalties: int or None,
            season: Season or str or int,
            competition: Competition or str or int,
            venue: SportVenue or str or int,
            status: str,
            url: Url,
            allow_overwrite: bool = False,
    ):
        fixture = Fixture(
            id=None,
            external_id=external_id,
            start_at=start_at,
            is_fulltime=is_fulltime,
            is_terminated=is_terminated,
            home_id=None,
            away_id=None,
            home_score={'goals': home_goals, 'penalties': home_penalties},
            away_score={'goals': away_goals, 'penalties': away_penalties},
            season_id=None,
            competition_id=None,
            venue_id=None,
            status=status,
            url=url,
            allow_overwrite=allow_overwrite
        )
        if isinstance(home, (str, int)):
            home = Team.placeholder(id=home)
        fixture._home = home
        if isinstance(away, (str, int)):
            away = Team.placeholder(id=away)
        fixture._away = away
        if isinstance(season, (str, int)):
            season = Season.placeholder(id=season)
        fixture._season = season
        if isinstance(competition, (str, int)):
            competition = Competition.placeholder(id=competition)
        fixture._competition = competition
        if isinstance(venue, (str, int)):
            venue = SportVenue.placeholder(id=venue)
        fixture._venue = venue
        return fixture

    @property
    def id(self):
        return self._id

    @property
    def competition_id(self):
        return self._competition.id

    @property
    def home_id(self):
        return self._home.id

    @property
    def away_id(self):
        return self._away.id

    @property
    def home(self):
        return self._home

    @property
    def away(self):
        return self._away

    @property
    def season_id(self):
        return self._season.id

    @property
    def venue_id(self):
        return self._venue.id

    @property
    def home_score(self):
        return Score(
            goals=self._home_goals,
            penalties=self._home_penalties
        )

    @property
    def home_goals(self):
        return self._home_goals

    @home_goals.setter
    def home_goals(self, goals):
        self._home_goals = goals

    @property
    def away_score(self):
        return Score(
            goals=self._away_goals,
            penalties=self._away_penalties
        )

    @property
    def away_goals(self):
        return self._away_goals

    @away_goals.setter
    def away_goals(self, goals):
        self._away_goals = goals

    @property
    def status(self):
        return self._status

    @property
    def is_fulltime(self):
        return self._is_fulltime

    @property
    def is_terminated(self):
        return self._is_terminated

    @property
    def start_at(self):
        return self._start_at

    @property
    def external_id(self):
        return self._external_id

    @property
    def home_penalties(self):
        return self._home_penalties

    @property
    def away_penalties(self):
        return self._away_penalties

    def __repr__(self):
        return '<Fixture(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'competition_id', 'home_id', 'away_id', 'season_id', 'venue_id',
            'home_score', 'away_score', 'status', 'is_fulltime', 'start_at', 'external_id', 'url',
            'allow_overwrite', 'is_terminated'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences




class MagicField:
    def __init__(
            self,
            id: int or None,
            key: str,
            value: str,
            user_id: str or int,
    ):
        from livestyled.models.user import User
        self.id = id
        self.key = key
        self.value = value
        self._user = User.placeholder(id=user_id)

    @classmethod
    def create_new(
            cls,
            user,
            key: str,
            value: str,
    ):
        magic_field = MagicField(
            id=None,
            key=key,
            value=value,
            user_id=None,
        )
        from livestyled.models.user import User
        if isinstance(user, (str, int)):
            user = User.placeholder(id=user)
        magic_field._user = user
        return magic_field

    @property
    def user_id(self):
        return self._user.id

    @property
    def user(self):
        return self._user

    def __repr__(self):
        return '<MagicField(id={self.id!r}, key={self.key!r}, value={self.value!r}, user={self.user!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'key', 'value'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.competition import Competition


class CompetitionSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'competitons'
        url = 'v4/competitions'
        model = Competition

    id = fields.Int()
    label = fields.String()
    name = fields.String()
    status = fields.String(allow_none=True)
    external_id = fields.String(data_key='externalId', missing=None)
    logo_url = fields.String(data_key='logoUrl', allow_none=True)
    sort_id = fields.Int(data_key='sortId')



class App:
    def __init__(
            self,
            id,
            currency,
            email,
            roles,
            password,
            salt,
            username,
            code,
            token,
            name,
            api_keys,
            timezone,
            google_api_key,
            payment_gateway,
            venues,
            sender_email,
            deeplink_namespace,
            merchant_accounts,
            title,
            status,
            cohorts
    ):
        self.id = id
        if currency:
            if isinstance(currency, Currency):
                self.currency = currency
            elif isinstance(currency, dict):
                self.currency = Currency(**currency)
        else:
            self.currency = None


class Currency:
    def __init__(self, id, title, iso_code, sign):
        self.id = id
        self.title = title
        self.iso_code = iso_code
        self.sign = sign



class Competition:
    def __init__(
            self,
            id: int or None,
            name: str or None,
            status: str or None,
            external_id: str or None,
            logo_url: str or None,
            sort_id: int or None,
    ):
        self.id = id
        self.name = name
        self.status = status
        self.logo_url = logo_url
        self.sort_id = sort_id
        self.external_id = external_id

    @classmethod
    def create_new(
            self,
            name: str,
            status: str or None,
            external_id: str or None,
            logo_url: str or None,
            sort_id: int,
    ):
        self.id = None
        self.name = name
        self.status = status
        self.logo_url = logo_url
        self.sort_id = sort_id
        self.external_id = external_id
        return self

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            name=None,
            status=None,
            external_id=None,
            logo_url=None,
            sort_id=None
        )

    def diff(self, other):
        differences = {}
        for key, value in self.__dict__.items() - other.__dict__.items():
            differences[key] = value
        return differences



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.team import Team


class TeamSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'teams'
        url = 'v4/teams'
        model = Team

    id = fields.Int(required=False, allow_none=False)
    name = fields.String(data_key='name')
    short_name = fields.String(data_key='shortName')
    light_crest_url = fields.String(data_key='lightCrestUrl', missing=None)
    dark_crest_url = fields.String(data_key='darkCrestUrl', missing=None)
    external_id = fields.String(data_key='externalId', missing=None)



from datetime import datetime, timedelta, timezone
import os

from livestyled.models.device import Device
from livestyled.models.device_preference import DevicePreference
from livestyled.models.event import Event
from livestyled.models.venue import Venue
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_device_preferences(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/device_preferences', 'mock_responses/ls_api/device_preferences_page1.json', 200),
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/device_preferences?page=2', 'mock_responses/ls_api/device_preferences_page2.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    device_preferences = list(resource_client.get_device_preferences())
    assert device_preferences
    assert len(device_preferences) == 31
    device_preference = device_preferences[0]
    assert isinstance(device_preference, DevicePreference)
    assert device_preference.venue
    assert isinstance(device_preference.venue, Venue)
    assert device_preference.venue.id == 10000918
    assert device_preference.id == 1
    assert device_preference.device
    assert isinstance(device_preference.device, Device)
    assert device_preference.device.id == 1620610
    assert not device_preference.event
    assert device_preference.created_at == datetime(2019, 2, 12, 15, 12, 46, tzinfo=timezone(timedelta(0), '+0000'))

    device_preference = device_preferences[3]
    assert isinstance(device_preference, DevicePreference)
    assert device_preference.id == 4
    assert not device_preference.venue
    assert device_preference.device
    assert isinstance(device_preference.device, Device)
    assert device_preference.device.id == 1748339
    assert device_preference.event
    assert isinstance(device_preference.event, Event)
    assert device_preference.event.id == 100015921
    assert device_preference.created_at == datetime(2019, 8, 27, 17, 19, 6, tzinfo=timezone(timedelta(0), '+0000'))


def test_get_device_preferences_default_ordering(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/device_preferences', 'mock_responses/ls_api/device_preferences_page1.json', 200),
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/device_preferences?page=2', 'mock_responses/ls_api/device_preferences_page2.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    list(resource_client.get_device_preferences())

    assert len(requests_mock.request_history) == 2
    assert requests_mock.request_history[0].url == 'https://test.livestyled.com/v4/device_preferences?order%5BcreatedAt%5D=desc'



from datetime import datetime, timedelta, timezone
import os

from livestyled.models.event import Event
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_events(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/events', 'mock_responses/ls_api/events_page1.json', 200),
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/events?page=2', 'mock_responses/ls_api/events_page2.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    events = list(resource_client.get_events())
    assert events
    assert len(events) == 31
    event = events[0]
    assert isinstance(event, Event)
    assert event.title == 'SABATON'
    assert event.status == 'ACTIVE'
    assert event.description == "Thirty-three drivers. 200 laps. 500 miles. One bottle of cold milk. The only thing missing from this magical Indianapolis 500 mix is you. Don't miss out on another incredible edition of \"The Greatest Spectacle in Racing.\""
    assert event.image_url == 'http://d255vb63773d25.cloudfront.net/-/media/IMS/events/schedule-images/indy-500/2020/Indy500.png?vs=1&d=20200326T144925Z'
    assert event.promoted is False
    assert event.created_at == datetime(2015, 7, 22, 17, 53, 56, tzinfo=timezone(timedelta(0), '+0000'))
    assert event.updated_at == datetime(2020, 4, 9, 15, 0, 42, tzinfo=timezone(timedelta(0), '+0000'))
    assert len(event.event_dates) == 1
    assert event.event_dates[0].id == 5448
    assert event.event_dates[0].start_at == datetime(2017, 3, 29, 16, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert event.event_dates[0].end_at == datetime(2017, 3, 29, 19, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert event.event_dates[0].general_ticket_url == 'https://www.ticketmaster.co.uk/Blue-tickets/artist/1843366'
    assert event.translations is None
    assert len(event.venues) == 1
    assert event.venues[0].id == 10000994



from datetime import datetime, timedelta, timezone
import os

from livestyled.models import News
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_news_by_id(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/news/1234', 'mock_responses/ls_api/news_1234.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    news = resource_client.get_news_article(1234)
    assert isinstance(news, News)
    assert news.id == 1234
    assert news.title == 'Things happened'
    assert news.external_id == '50990'
    assert news.published_at == datetime(2020, 2, 23, 4, 56, 59, tzinfo=timezone(timedelta(0), '+0000'))
    assert news.headline == 'Some news'
    assert news.author is None
    assert news.updated_at == datetime(2020, 3, 18, 1, 15, 7, tzinfo=timezone(timedelta(0), '+0000'))
    assert news.media == {
        'type': 'EXTERNALVIDEO',
        'url': 'https://www.thenews.com/video',
        'thumbnail_url': None,
    }


def test_get_news(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/news', 'mock_responses/ls_api/news.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    news = resource_client.get_news_articles()
    news = [n for n in news]
    assert len(news) == 2
    news.sort(key=lambda n: n.id)
    assert news[0].id == 1234
    assert news[1].id == 1235



from livestyled.models.event import Event
from livestyled.models.ticket_integration import TicketIntegration
from livestyled.models.user import User
from livestyled.models.venue import Venue


class Ticket:
    def __init__(
            self,
            id,
            external_ticket_id,
            seat,
            qr_code_url,
            title,
            external_event_id,
            barcode,
            sector_name,
            venue_name,
            venue_room,
            client_name,
            premium,
            client_email,
            price,
            share_link,
            external_customer_ref,
            entrance,
            section,
            row,
            price_code,
            created_at,
            updated_at,
            user_id,
            status,
            session_date=None,
            can_share=False,
            sharer_email=None,
            sharer_id=None,
            redeemed_at=None,
            redeemer_id=None,
            share_code=None,
            redeemer_email=None,
            parent_ticket=None,
            shared_at=None,
            legal_long_text=None,
            legal_short_text=None,
            map_url=None,
            map_image_url=None,
            ticket_integration=None,
            venue=None,
            event=None,
    ):
        self.id = id
        self.external_ticket_id = external_ticket_id
        self.seat = seat
        self.qr_code_url = qr_code_url
        self.session_date = session_date
        self.title = title
        self.external_event_id = external_event_id
        self.barcode = barcode
        self.sector_name = sector_name
        self.venue_name = venue_name
        self.venue_room = venue_room
        self.client_name = client_name
        self.premium = premium
        self.client_email = client_email
        self.price = price
        self.created_at = created_at
        self.updated_at = updated_at
        self.share_link = share_link
        self.external_customer_ref = external_customer_ref
        self.entrance = entrance
        self.section = section
        self.row = row
        self.price_code = price_code
        if user_id:
            self._user = User.placeholder(id=user_id)
        else:
            self._user = None
        self.status = status
        self.can_share = can_share
        self.sharer_email = sharer_email
        self.redeemed_at = redeemed_at
        self.share_code = share_code
        self.redeemer_email = redeemer_email
        self.shared_at = shared_at
        if sharer_id:
            self._sharer = User.placeholder(id=sharer_id)
        else:
            self._sharer = None
        if redeemer_id:
            self._redeemer = User.placeholder(id=redeemer_id)
        else:
            self._redeemer = None
        if parent_ticket:
            if isinstance(parent_ticket, dict):
                self._parent_ticket = Ticket(**parent_ticket)
            elif isinstance(parent_ticket, (int, str)):
                self._parent_ticket = Ticket.placeholder(id=int(parent_ticket))
        else:
            self._parent_ticket = None
        self.legal_long_text = legal_long_text
        self.legal_short_text = legal_short_text
        self.map_url = map_url
        self.map_image_url = map_image_url
        if ticket_integration:
            if isinstance(ticket_integration, dict):
                self._ticket_integration = TicketIntegration(**ticket_integration)
            elif isinstance(ticket_integration, (int, str)):
                self._ticket_integration = TicketIntegration.placeholder(ticket_integration)
            elif isinstance(ticket_integration, TicketIntegration):
                self._ticket_integration = ticket_integration
        else:
            self._ticket_integration = None

        if event:
            if isinstance(event, Event):
                self.event = event
            elif isinstance(event, (int, str)):
                self.event = Event.placeholder(id=event)
            elif isinstance(event, dict):
                self.event = Event(**event)
        else:
            self.event = None

        if venue:
            if isinstance(venue, Venue):
                self.venue = venue
            elif isinstance(venue, (int, str)):
                self.venue = Venue.placeholder(id=venue)
            elif isinstance(venue, dict):
                self.venue = Venue(**venue)
        else:
            self.venue = None

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            external_ticket_id=None,
            seat=None,
            qr_code_url=None,
            title=None,
            external_event_id=None,
            barcode=None,
            sector_name=None,
            venue_name=None,
            venue_room=None,
            client_name=None,
            premium=None,
            client_email=None,
            price=None,
            share_link=None,
            external_customer_ref=None,
            entrance=None,
            section=None,
            row=None,
            price_code=None,
            created_at=None,
            updated_at=None,
            user_id=None,
            status=None,
            session_date=None,
            can_share=False,
            sharer_email=None,
            sharer_id=None,
            redeemed_at=None,
            redeemer_id=None,
            share_code=None,
            redeemer_email=None,
            parent_ticket=None,
            shared_at=None,
            legal_long_text=None,
            legal_short_text=None,
            map_url=None,
            map_image_url=None,
            ticket_integration=None,
            venue=None,
            event=None
        )

    @classmethod
    def create_new(
            cls,
            user: User or str or int,
            external_ticket_id=None,
            seat=None,
            qr_code_url=None,
            session_date=None,
            title=None,
            external_event_id=None,
            barcode=None,
            sector_name=None,
            venue_name=None,
            venue_room=None,
            client_name=None,
            premium=False,
            client_email=None,
            price=0,
            row=None,
            section=None,
            share_link=None,
            external_customer_ref=None,
            price_code=None,
            entrance=None,
            status=None,
            can_share=False,
            sharer_email=None,
            sharer: User or str or int or None = None,
            redeemed_at=None,
            redeemer: User or str or int or None = None,
            share_code=None,
            redeemer_email=None,
            parent_ticket=None,
            shared_at=None,
            legal_long_text=None,
            legal_short_text=None,
            map_url=None,
            map_image_url=None,
            ticket_integration=None,
            venue: Venue or str or int or None = None,
            event: Event or str or int or None = None,
    ):
        ticket = Ticket(
            id=None,
            external_ticket_id=external_ticket_id,
            seat=seat,
            qr_code_url=qr_code_url,
            session_date=session_date,
            title=title,
            external_event_id=external_event_id,
            barcode=barcode,
            sector_name=sector_name,
            venue_name=venue_name,
            venue_room=venue_room,
            client_name=client_name,
            premium=premium,
            client_email=client_email,
            price=price,
            created_at=None,
            updated_at=None,
            share_link=share_link,
            external_customer_ref=external_customer_ref,
            entrance=entrance,
            section=section,
            row=row,
            price_code=price_code,
            user_id=None,
            status=status,
            can_share=can_share,
            sharer_email=sharer_email,
            sharer_id=None,
            redeemed_at=redeemed_at,
            redeemer_id=None,
            share_code=share_code,
            redeemer_email=redeemer_email,
            parent_ticket=None,
            shared_at=shared_at,
            legal_long_text=legal_long_text,
            legal_short_text=legal_short_text,
            map_url=map_url,
            map_image_url=map_image_url,
            ticket_integration=ticket_integration,
            venue=venue,
            event=event,
        )
        if isinstance(user, (str, int)):
            user = User.placeholder(id=user)
        ticket._user = user
        if isinstance(sharer, (str or int)):
            sharer = User.placeholder(id=sharer)
        ticket._sharer = sharer
        if isinstance(redeemer, (str or int)):
            redeemer = User.placeholder(id=redeemer)
        ticket._redeemer = redeemer
        if isinstance(parent_ticket, (str or int)):
            parent_ticket = Ticket.placeholder(id=parent_ticket)
        ticket._parent_ticket = parent_ticket
        return ticket

    @property
    def user_id(self):
        if self._user:
            return self._user.id
        else:
            return None

    @property
    def user(self):
        return self._user

    @property
    def redeemer_id(self):
        if self._redeemer:
            return self._redeemer.id
        else:
            return None

    @property
    def redeemer(self):
        return self._redeemer

    @property
    def sharer_id(self):
        if self._sharer:
            return self._sharer.id
        else:
            return None

    @property
    def sharer(self):
        return self._sharer

    @property
    def parent_ticket(self):
        return self._parent_ticket

    @property
    def ticket_integration(self):
        return self._ticket_integration

    def __repr__(self):
        return '<Ticket(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'external_ticket_id', 'seat', 'qr_code_url', 'session_date', 'title', 'external_event_id',
            'barcode', 'sector_name', 'venue_name', 'venue_room', 'client_name', 'premium', 'client_email',
            'price', 'status', 'can_share', 'sharer_email', 'redeemed_at', 'redeemer_id', 'share_code',
            'redeemer_email', 'parent_ticket', 'shared_at', 'legal_long_text', 'legal_short_text', 'map_url',
            'map_image_url', 'ticket_integration', 'entrance', 'row', 'section', 'price_code', 'external_customer_ref',
            'venue', 'event'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



class TicketIntegration:
    def __init__(
            self,
            id,
            label,
            name,
            adapter,
            config_payload,
            endpoint_url,
            auth_required,
            module,
            login_request,
            default,
            can_share
    ):
        self.id = id
        self.label = label
        self.name = name
        self.adapter = adapter
        self.config_payload = config_payload
        self.endpoint_url = endpoint_url
        self.auth_required = auth_required
        self.module = module
        self.login_request = login_request
        self.default = default
        self.can_share = can_share

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            label=None,
            name=None,
            adapter=None,
            config_payload=None,
            endpoint_url=None,
            auth_required=None,
            module=None,
            login_request=None,
            default=None,
            can_share=None
        )

    @classmethod
    def create_new(
            cls,
            endpoint_url,
            login_request,
            adapter,
            label=None,
            name=None,
            config_payload=None,
            auth_required=None,
            module=None,
            default=None,
            can_share=None
    ):
        ticket_integration = TicketIntegration(
            id=None,
            label=label,
            name=name,
            adapter=adapter,
            config_payload=config_payload,
            endpoint_url=endpoint_url,
            auth_required=auth_required,
            module=module,
            login_request=login_request,
            default=default,
            can_share=can_share
        )
        return ticket_integration

    def __repr__(self):
        return '<TicketIntegration(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'label', 'name', 'endpoint_url', 'adapter', 'config_payload',
            'auth_required', 'module', 'login_request', 'default', 'can_share'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences
